#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdint.h>
#include <assert.h> 
#include <string.h>
#include "tsadclib1624.h"

extern char *optarg;
extern int optind, opterr, optopt;

bool getToken(uint8_t *theToken, uint8_t **args, uint8_t *delimiter);

#define MaxRangeString 32
#define MaxSampleRate 1000000
#define DefaultSampleRate 10000

#define MaxBufferSize 100000
#define MaxChannels 24
uint16_t *bufferArray[MaxChannels];
int16_t *workArray[MaxChannels];

typedef enum {
    PF_NONE,
    PF_DECIMAL,
    PF_HEX,
    PF_CSV
} printFormat;

/*
 *  parsechannels
 *
 *  Parse the options string and return a 32 bit value with a bit set for
 *  each ADC channel of interest
 */

uint32_t parsechannels(uint8_t *optionsString, uint16_t maximumChannel)
{
    uint32_t returnChannels = 0;
    uint16_t theBit;
    uint8_t theDelimiter;
    uint8_t *myArgs    = optionsString;
    uint8_t *errorArgs = optionsString;
    uint8_t theChannelString[80];
    int32_t theChannel = -1;
    int32_t theHighChannel = -1;


    //Determine which channels to sample
    while (1)
    {
        if (!getToken(theChannelString, &myArgs, &theDelimiter))
            return 0;  // String token is invalid here

        if ((theChannelString[1] == '\0' && isdigit(theChannelString[0])) ||
                (theChannelString[2] == '\0' && isdigit(theChannelString[0]) && isdigit(theChannelString[1])))
        {
            theChannel = strtol(theChannelString, NULL, 10);
            if (theChannel >= maximumChannel)
            {
                printf("Invalid ADC channel: %d (Range: 0-%d)\n", theChannel, maximumChannel);
                return 0;
            }
        }
        else
        {
            fprintf(stderr, "Invalid format, Sample ADC channels CHANS, e.g. \"1-3,5\", \"2,3,5\"\n", errorArgs, myArgs);
            return 0;
        }

        if (theDelimiter == '-')
        {
            theHighChannel = theChannel;

            if (!getToken(theChannelString, &myArgs, &theDelimiter))
                return 0;  // String token is invalid here

            if ((theChannelString[1] == '\0' && isdigit(theChannelString[0])) ||
                    (theChannelString[2] == '\0' && isdigit(theChannelString[0]) && isdigit(theChannelString[1])))
            {
                theHighChannel = strtol(theChannelString, NULL, 10);
                if (theHighChannel >= maximumChannel)
                {
                    printf("Invalid ADC channel: %d  (Range: 0-%d)\n", theChannel, maximumChannel - 1);
                    return 0;
                }
            }
            else
            {
                fprintf(stderr, "Invalid format, Sample ADC channels CHANS, e.g. \"1-3,5\", \"2,3,5\"\n", errorArgs, myArgs);
                return 0;
            }
        }
        else if ((theDelimiter != ',') && (theDelimiter != ' ' && theDelimiter != '\0'))
        {
            fprintf(stderr, "Invalid format, Sample ADC channels CHANS, e.g. \"1-3,5\", \"2,3,5\"\n", errorArgs, myArgs);
            return 0;
        }

        //  Assign bits per channel
        if (theHighChannel != -1)
            for (theBit = theChannel; theBit <= theHighChannel; theBit++)
                returnChannels |= (1 << theBit);
        else
            returnChannels |= (1 << theChannel);

        if (theDelimiter == ' ' || theDelimiter == '\0')
            break;

        theChannel = -1;
        theHighChannel = -1;
    }

    return returnChannels;
}

void usage(char **argv)
{
    fprintf(stderr, "\nUsage: tsadc1624ctl -m model [OPTION] ...\n"
            "Interact with the ADC-16 and ADC-24 in the 7200, 7350, 7370, & 7800 SBCs\n"
            "General options:\n"
            "  -a  channels*        Acquire ADC channels (1 or 2), e.g. \"1-2\", \"3,4\"\n" 
            "                       Note: you must always specify an odd channel and\n"
            "                       the subsequent even channel, e.g., \'1-2\' or \'7,8\'\n"
            "  -b  board*           ADC board number (0-3) on whici to sample\n"
            "                       Required prior to \'r\', \'s\', \'d\' and \'D\' options\n"
            "  -B                   Maximum buffer size\n"
            "  -D                   Set DAC output - Range\n"
            "  -d                   Set DAC otuput - Channel:Value\n"
            "  -e                   Enumerate the list of installed ADC boards and quit\n"
            "  -h                   This help screen\n"
            "  -m  model (Required) Model of the processor (m modeal [from list above])\n"
            "                       NOTE: This parameter must be first on the line\n"
            "  -M                   Maximum channels (Range 1-16 OR 1-24)\n"
            "  -n  number*          Number of samples to acquire (Default: 10000)\n"
            "  -o  format           printf output to stdout in format: decimal, hex, csv\n"			
            "  -p                   Utilize non-blocking Polling\n"
            "  -r  rate*            Sample rate in Hz (Default: 10000)\n" 
            "  -R                   display details on range selectionz\n" 
            "  -s  channels         Sample ADC channels CHANS, e.g. \"1-3,5\", \"2,3,5\"\n" 
            "                       output string parseable data to standard out\n"
            "  -v                   Verbose output\n"			
            "  -x  range*           Range and measurement mode (Refer to enum \'adcrange\' in the library header file\n" 
            "                       ADC16: 55S, 05S (default), 1010S, 0010S, 55D, 05D, 1010D, 0010D\n" 
            "                       ADC24: 0VS, 02VS (default), 02VAS, 02VBS, 0VD, 02VD, 02VAD, 02VBD\n" 
            "  -?                   This help screen\n"
            "   *                   Option required for \'a\' and \'s\'\n");
}

const char *optionsList = "vehpR?m:a:s:b:n:r:x:M:B:o:D:d:";
sbctype boardModel = SBC_NONE;
adctype adcBoardStatus;

typedef struct {
    uint8_t *theName;
    adcrange theRange;
} adcRangeSelection;

adcRangeSelection rangeTable16[] = {
    { "55S",   ADCRANGE_55S },
    { "05S",   ADCRANGE_05S },
    { "1010S", ADCRANGE_1010S },
    { "0010S", ADCRANGE_0010S },
    { "55D",   ADCRANGE_55D },
    { "05D",   ADCRANGE_05D },
    { "1010D", ADCRANGE_1010D },
    { "0010D", ADCRANGE_0010D },
    { NULL, 0 }
};

adcRangeSelection rangeTable24[] = {
    { "0VS",   ADCRANGE_0VS },
    { "02VS",  ADCRANGE_02VS },
    { "02VAS", ADCRANGE_02VAS },
    { "02VBS", ADCRANGE_02VBS },
    { "0VD",   ADCRANGE_0VD },
    { "02VD",  ADCRANGE_02VD },
    { "02VAD", ADCRANGE_02VAD },
    { "02VBD", ADCRANGE_02VBD },
    { NULL, 0 }
};

typedef struct {
    uint8_t *theName;
    adcrange theRange;
} dacRangeSelection;

dacRangeSelection rangeTableDac[] = {
    { "0VU",  DACRANGE_0VU },
    { "02VU", DACRANGE_02VU },
    { "0VB",  DACRANGE_0VB },
    { "02VB", DACRANGE_02VB },
    { NULL, 0 }
};

bool bFlag = false;  // Board number
bool dFlag = false;  // DAC Output
bool DFlag = false;  // DAC Output
bool eFlag = false;  // Enumerate
bool mFlag = false;  // Model (SBC)
bool nFlag = false;  // Number of samples
bool oFlag = false;  // Output (printing) type
bool pFlag = false;  // Polling enabled
bool rFlag = false;  // Rate to sample
bool vFlag = false;  // Verbise
bool xFlag = false;  // The range

uint16_t maximumChannels          = 24;
uint32_t maximumBufferSize        = 10000;
uint32_t numberOfSamplesToAcquire = 10000;

void outputWithFormat(uint16_t theChannel, int16_t *sPtr, uint32_t sampleCnt, printFormat type)
{
    uint32_t index;

    switch (type)
    {
        case PF_DECIMAL:
            {
                for (index = 0; index < sampleCnt; index++)
                {
                    if ((index % 10) == 0)
                        printf("\n%02d:%6d: ", theChannel, index);
                    printf("%5d ", *sPtr++);
                }
                break;
            }

        case PF_HEX:
            {
                for (index = 0; index < sampleCnt; index++)
                {
                    if ((index % 8) == 0)
                        printf("\n%02d:%06x: ", theChannel, index);
                    printf("0x%04x ", (uint16_t) *sPtr++);
                }
                break;
            }

        case PF_CSV:
            printf("\n%d,", theChannel);
            for (index = 0; index < sampleCnt; index++)
            {
                if ((index % 10) == 0)
                    printf("\n");
                printf("%d,", *sPtr++);
            }
            break;

        case PF_NONE:
        default:
            break;
    }
}

int main(int argc, char **argv)
{
    int aChar;
    uint16_t length;
    uint16_t index;
    uint16_t channelCount;
    int32_t numberOfSamplesAcquired;
    int returnStatus;
    uint32_t channelsToSample;
    BoardIndex anIndex;
    uint16_t maxChannels = 16;
    int16_t whichBoard = 0;
    uint16_t *aBuffer;
    uint8_t tokenString[81];
    uint8_t operation = '\0';
    uint8_t whichChannel = 0;
    uint32_t adcSampleRate = DefaultSampleRate;
    uint8_t adcRangeString[MaxRangeString+1];
    uint16_t adcRangeValue = 0;
    adcRangeSelection *theRange;
    dacRangeSelection *theDacRange;
    uint8_t dacRangeString[MaxRangeString+1];
    uint16_t dacRangeValue = 0;
    uint32_t dacCchannel, dacValue;
    printFormat outputType = PF_NONE;

    if (argc == 1)
    {
        usage(argv);
        return 1;
    }

    adcRangeString[0] = '\0';
    dacRangeString[0] = '\0';

    //  Build buffers
    for (index = 0; index < MaxChannels; index++)
    {
        if ((aBuffer = malloc(MaxBufferSize+1)) == NULL)
        {
            fprintf(stderr, "tsadc1624ctl can't allocate memory\n");
            return 1;
        }

        bufferArray[index] = aBuffer;
    }

    //EWD: If time remains modify me to getopt_long()
    while ((aChar = getopt(argc, argv, optionsList)) != -1)
    {
        //printf("Option: \'%c\'  Optarg: %s\n", aChar, optarg);
        switch(aChar)
        {
            case 'b':  // Set board number and details
                if ((whichBoard = strtoul(optarg, NULL, 10)) >= ADC_BOARDMAX)
                {
                    fprintf(stderr, "Invalid ADC board number: %s  (Range: %d to %d)\n",
                            ADC_BOARD0, ADC_BOARDMAX-1, optarg);
                    usage(argv);
                    return 1;
                }

                bFlag = true;
                break;

            case 'd':  // Set DAC channel & value
                {
                    char *delimiter;

                    if ((dacCchannel = strtoul(optarg, &delimiter, 10)) > 3)
                    {
                        fprintf(stderr, "Invalid DAC channel number: %s  (Range: 0 to 3)\n", dacCchannel);
                        usage(argv);
                        return 1;
                    }

                    if (*delimiter != ':')
                    {
                        fprintf(stderr, "Invalid DAC channel delimiter.  channel:value\n");
                        usage(argv);
                        return 1;
                    }

                    if ((dacValue = strtoul(++delimiter, NULL, 10)) > 0xffff)
                    {
                        fprintf(stderr, "Invalid DAC value: %d\n", dacValue);
                        usage(argv);
                        return 1;
                    }

                    dFlag = true;
                }
                break;


            case 'D':  // Set DAC RABGE
                {
                    if (strlen(optarg) < MaxRangeString)
                        strcpy(dacRangeString, optarg);
                    else
                    {
                        fprintf(stderr, "Invalid DAC range string: %s\n", optarg);
                        usage(argv);
                        return 1;
                    }
                    //  Setup range

                    theDacRange = rangeTableDac;

                    while (theDacRange->theName != NULL)
                    {
                        if (strcmp(theDacRange->theName, dacRangeString) == 0)
                        {
                            dacRangeValue = theDacRange->theRange;
                            break;
                        }
                        else
                            theDacRange++;
                    }

                    if (theDacRange->theName == NULL) // Ran out of table
                    {
                        fprintf(stderr, "Invalid DAC range specified: \'%s\' - See \'-R\' option\n", dacRangeString);
                        usage(argv);
                        return -1;
                    }

                    DFlag = true;
                }
                break;

            case 'o':  // Set output type
                if (strncmp(optarg, "dec", 3) == 0)
                    outputType = PF_DECIMAL;
                else if (strncmp(optarg, "hex", 3) == 0)
                    outputType = PF_HEX;
                else if (strncmp(optarg, "csv", 3) == 0)
                    outputType = PF_CSV;
                else
                {
                    fprintf(stderr, "Invalid print output type: %s\n", optarg);
                    usage(argv);
                    return 1;
                }
                oFlag = true;
                break;

            case 'm':  // Set board model
                if (strcmp(optarg, "7200") == 0)
                    boardModel = SBC_TS7200;
                else if (strcmp(optarg, "7350") == 0)
                    boardModel = SBC_TS7350;
                else if (strcmp(optarg, "7370") == 0)
                    boardModel = SBC_TS7350;
                else if (strcmp(optarg, "7800") == 0)
                    boardModel = SBC_TS7800;
                else
                {
                    fprintf(stderr, "Invalid SBC board model ID: %s\n", optarg);
                    usage(argv);
                    return 1;
                }

                printf("SBC Model assigned as: %s\n", optarg);
                mFlag = true;
                break;

            case 'e':  // Initialize and enumerate boards
                eFlag = true;
                break;

            case 'x':
                if (strlen(optarg) < MaxRangeString)
                    strcpy(adcRangeString, optarg);
                else
                {
                    fprintf(stderr, "Invalid range string: %s\n", optarg);
                    usage(argv);
                    return 1;
                }
                xFlag = true;
                break;

            case 'B':
                maximumBufferSize = strtoul(optarg, NULL, 10);
                if (maximumBufferSize == 0)
                {
                    fprintf(stderr, "Maximum buffer size range is 1-%d and you chose: %d\n", MaxBufferSize, maximumBufferSize);
                    usage(argv);
                    return 1;
                }
                else if (maximumBufferSize > MaxBufferSize)
                {
                    fprintf(stderr, "Setting requested maximum buffer size (%d) to: %d\n", maximumBufferSize, MaxBufferSize);
                    maximumBufferSize = MaxBufferSize;
                }
                break;

            case 'M':
                maximumChannels = strtoul(optarg, NULL, 10);
                if (maximumChannels == 0 || maximumChannels > 24)
                {
                    fprintf(stderr, "Maximum channels range is 1-24 and you chose: %d\n", maximumChannels);
                    usage(argv);
                    return 1;
                }
                break;

            case 'n':
                numberOfSamplesToAcquire = strtoul(optarg, NULL, 10);
                if (numberOfSamplesToAcquire > MaxBufferSize || numberOfSamplesToAcquire == 0)
                {
                    fprintf(stderr, "Number of samples range is 1-%d and you chose: %d\n", MaxBufferSize, numberOfSamplesToAcquire);
                    usage(argv);
                    return 1;
                }
                nFlag = true;
                break;

            case 'r':
                adcSampleRate = strtoul(optarg, NULL, 10);
                if (adcSampleRate > MaxSampleRate || adcSampleRate == 0)
                {
                    fprintf(stderr, "Maximum sample rate is %d and you chose: %d\n", MaxSampleRate, adcSampleRate);
                    usage(argv);
                    return 1;
                }
                rFlag = true;
                break;

            case 'p':
                pFlag = true;
                break;

            case 'a':
                operation = 'a';
                // NOTE: Falls through here

            case 's':
                if (operation == '\0')
                    operation = 's';

                //  Process channels to sample
                for (length = 0; length < 80; length++)
                    if (optarg[length] == '\0')
                        break;
                strncpy(tokenString, optarg, length);
                tokenString[length] = '\0';
                if ((channelsToSample = parsechannels(tokenString, 24)) == 0)
                    return -1;
                break;

            case 'v':
                printf("Setting verbose\n");
                vFlag = true;
                tsadclib1624_verbose(true);
                break;

            case 'R':	
                fprintf(stderr, "ADC and DAC range definitions\n"
                        "ADC16\n"
                        "55S     -5 to +5V  single ended\n"
                        "05S      0 to +5V  single ended (default)\n"
                        "1010S  -10 to +10V single ended\n"
                        "0010S    0 to +10V single ended\n"
                        "55D     -5 to +5V  differential\n"
                        "05D      0 to +5V  differential\n"
                        "1010D  -10 to +10V differential\n"
                        "0010D    0 to +10V differential\n\n"
                        "ADC24\n"
                        "0VS     0 to +VRef   single ended\n"
                        "02VS    0 to +2xVRef single ended (default)\n"
                        "02VAS   0 to +2xVRef single ended\n"
                        "02VBS   0 to +2xVRef single ended\n"
                        "0VD     0 to +VRef   differential\n"
                        "02VD    0 to +2xVRef differential\n"
                        "02VAD   0 to +2xVRef differential\n"
                        "02VBD   0 to +2xVRef differential\n\n"
                        "DAC\n"
                        "0VU     0 to VRef  - Unbuffered\n"
                        "02VU    0 to 2VRef - Unbuffered\n"
                        "0VB     0 to VRef  - Buffered\n"
                        "02VB    0 to 2VRef - Buffered\n");
                return 1;

            case 'h':	
            case '?':	
            default:
                usage(argv);
                return 1;
        }
    }

    if (maximumBufferSize < numberOfSamplesToAcquire)
    {
        fprintf(stderr, "Maximum buffer size (%d) < Number of samples (%d)\n", maximumBufferSize, numberOfSamplesToAcquire);
        usage(argv);
        return 1;
    }

    if (!mFlag)
    {
        fprintf(stderr, "SBC board model not set (-m)\n");
        usage(argv);
        return 1;
    }

    if ((returnStatus = tsadclib1624_init(boardModel, maximumChannels, maximumBufferSize)) != 0)
    {
        fprintf(stderr, "Board failed initialization: %d (%s)\n", returnStatus, tsadclib1624_errors(returnStatus));
        return 1;
    }

    if (bFlag)  //  Ensure selected board is installed and working
    {
        adcBoardStatus = tsadclib1624_boardstatus(whichBoard);

        if (adcBoardStatus == TS_ADCINVALID)
        {
            fprintf(stderr, "Invalid ADC board status - No board installed in slot: %d\n", whichBoard);
            usage(argv);
            return 1;
        }

        switch (adcBoardStatus)
        {
            case TS_ADC16:
                maxChannels = 16;
                break;

            case TS_ADC24:
                maxChannels = 24;
                break;

            default:
                maxChannels = 0;
        }
    }

    //  Process DAC output
    if (dFlag && DFlag)
    {
        if (adcBoardStatus == TS_ADC16)
        {
            tsadclib16_dacset(whichBoard, dacCchannel, dacRangeValue, dacValue);
            fprintf(stderr, "Setting Board:Dac:Value (Range) %d:%d:%d (%s)\n",
                    whichBoard, dacCchannel, dacValue, rangeTableDac[dacRangeValue >> 12].theName);
            return;
        }
        else  //  No DAC on ADC24
        {
            fprintf(stderr, "DAC options \'d\' and \'D\' invalid on ADC24\n");
            usage(argv);
            return 1;
        }
    }

    if (eFlag)
    {
        if (boardModel == SBC_NONE)
        {
            fprintf(stderr, "Model ID not set\n");
            usage(argv);
            return 1;
        }

        printf("\n");
        for (anIndex = ADC_BOARD0; anIndex < ADC_BOARDMAX; anIndex++)
        {
            adcBoardStatus = tsadclib1624_boardstatus(anIndex);
            if (adcBoardStatus == TS_ADCINVALID)
                printf("Board %01d: Not installed\n", anIndex);
            else
                printf("Board %01d: %s\n", anIndex, adcBoardStatus == TS_ADC16 ? "TSADC16" : "TSADC24");
        }
        printf("\n");
        return 0;
    }

    //  Setup ADC board information
    adcBoardStatus = tsadclib1624_boardstatus(whichBoard);

    if (adcBoardStatus == TS_ADCINVALID)
    {
        fprintf(stderr, "Invalid ADC in board slot: %d - use \'-b\' option\n", whichBoard);
        usage(argv);
        return -1;
    }
    else if (adcBoardStatus == TS_ADC24)
    {
        if (adcRangeString[0] == '\0')
            strcpy(adcRangeString, rangeTable24[1].theName);
        maxChannels = 24;
    }
    else
    {
        if (adcRangeString[0] == '\0')
            strcpy(adcRangeString, rangeTable16[1].theName);
        maxChannels = 16;
    }

    if (!xFlag)
    {
        adcRangeValue = (maxChannels == 16) ? rangeTable16[1].theRange : rangeTable24[1].theRange;
    }
    else
    {
        //  Setup range
        theRange = maxChannels == 16 ? rangeTable16 : rangeTable24;

        while (theRange->theName != NULL)
            if (strcmp(theRange->theName, adcRangeString) == 0)
            {
                adcRangeValue = theRange->theRange;
                break;
            }
            else
                theRange++;

        if (adcRangeValue == 0)
        {
            fprintf(stderr, "Invalid range specified: \'%s\' - See \'-R\' option\n", adcRangeString);
            usage(argv);
            return -1;
        }
    }

    //  In case larger channels were entered
    if (maxChannels == 16)
        channelsToSample &= 0x0000ffff;

    if (!tsadclib1624_buffers())
    {
        fprintf(stderr, "No work buffers allocated\n");
        return -1;
    }

    if (operation == 'a')  // Read 1 or 2 channels
    {
        channelCount = 0;
        for (index = 0; index < MaxChannels; index++)
            if (channelsToSample & (1 << index))
                channelCount++;

        if (channelCount > 2)
        {
            fprintf(stderr, "Invalid channel count, you must select only 2 adjacent\n");
            fprintf(stderr, " even and odd channels, e.g., \'0-1\' or 6,7\n");
            return -1;
        }

        //  Check for proper adjacent channel pair
        for (index = 0; index < MaxChannels; index++)
            if (channelsToSample & (1 << index))
                break;

        if (index == MaxChannels ||
                ((index & 1) != 0) ||  // if even channel
                (channelsToSample & (1 << (index + 1))) == 0)
        {
            fprintf(stderr, "Channel %d can only be selected with channel %d\n",
                    index, index & 1 ? index - 1 : index+1);
            return 1;
        }

        whichChannel = index;

        for (index = 0; index < MaxChannels; index++)
            workArray[index] = NULL;

        workArray[0] = bufferArray[0];
        if (channelCount == 2)
            workArray[1] = bufferArray[1];

        // Start the acquisution
        tsadclib1624_startAcquisution(whichBoard, whichChannel,
                adcSampleRate, numberOfSamplesToAcquire, adcRangeValue,
                workArray[0], workArray[1], false);

        if (pFlag)
        {
            numberOfSamplesAcquired = 0;

            while (numberOfSamplesAcquired == 0)
            {
                numberOfSamplesAcquired = tsadclib1624_pollResults(whichBoard, false);
                usleep(400);
            }
        }
        else
        {
            numberOfSamplesAcquired = tsadclib1624_pollResults(whichBoard, true);
        }
    }
    else if (operation == 's')  // read 'n' channels
    {
        if (channelsToSample == 0)
        {
            printf("main: No channels to sample\n");
            return -1;
        }

        // Find highest channel
        for (index = MaxChannels; index > 0; index--)
        {
            if (channelsToSample & (1 << (index - 1)))
            {
                whichChannel = index - 1;
                break;
            }
        }

        //  setup buffers
        for (index = 0; index < MaxChannels; index++)
            workArray[index] = bufferArray[index];

        // Start the sample
        tsadclib1624_sample(whichBoard, whichChannel, adcSampleRate, numberOfSamplesToAcquire, adcRangeValue, workArray, false);

        printf("main: Await data completion %s\n", pFlag ? "with polling" : "");
        if (pFlag)
        {
            numberOfSamplesAcquired = 0;

            while (numberOfSamplesAcquired == 0)
            {
                numberOfSamplesAcquired = tsadclib1624_pollResults(whichBoard, false);
                usleep(25);
            }
        }
        else
        {
            numberOfSamplesAcquired = tsadclib1624_pollResults(whichBoard, true);
        }
    }
    else
    {
        fprintf(stderr, "Must enter an operation parameter: \'a\' or \'s\' or DAC \'d\' and \'D\'\n");
        usage(argv);
        return 1;
    }

    if (oFlag)
    {
        if (outputType != PF_NONE)
        {
            if (operation == 'a')  // Read 1 or 2 channels
            {
                outputWithFormat(whichChannel, workArray[0], numberOfSamplesToAcquire, outputType);
                outputWithFormat(whichChannel+1, workArray[1], numberOfSamplesToAcquire, outputType);
            }
            else if (operation == 's')  // Read 1 to n channels
            {
                uint16_t index;

                for (index = 0; index < 24; index++)
                {
                    if (workArray[index] != NULL && channelsToSample & (1 << index))
                        outputWithFormat(index, workArray[index], numberOfSamplesToAcquire, outputType);
                }
                printf("\n");
            }
        }
    }

    return 0;
}

//
//  Return the next atomic token in the string
//

bool getToken(uint8_t *theToken, uint8_t **args, uint8_t *delimiter)
{
    uint8_t *workArgs = *args;  //  Get argument string
    uint8_t aChar;
    uint8_t *aToken = theToken;
    bool isArg = false;
    bool TokenIsAString = FALSE;
    uint8_t *inargs = workArgs;

    if (aToken == NULL)
        return FALSE;

    *aToken = '\0';

    if (workArgs == NULL || *workArgs == '\0')
        return FALSE;

    //  Skip leading whitespace
    while (TRUE)
    {
        if ((aChar = *workArgs) == '\0')
            break;

        //   Check for string token
        if (aChar == '\"')
        {
            TokenIsAString = TRUE;
            workArgs++;
            break;
        }

        if (aChar == ' ' || aChar == '\t')
            workArgs++;
        else
            break;
    }

    //  Now extract the token
    while (TRUE)
    {
        aChar = *workArgs;
        workArgs++;

        if (TokenIsAString)
        {
            if (aChar == '\0' ||
                    aChar == '\"' ||
                    aChar == '\r' ||
                    aChar == '\n')
                break;
        }
        else
        {
            if (aChar == '\0' ||
                    aChar == ' '  ||
                    aChar == ','  ||
                    aChar == '-'  ||
                    aChar == '\t' ||
                    aChar == '\r' ||
                    aChar == '\n')
                break;
        }

        *aToken++ = aChar;
        *aToken = '\0';
        isArg = true;
    }

    //  Check for comment - return NULL token if found
    if (strncmp("//", theToken, 2) == 0)
    {
        isArg = false;
        aToken = '\0';
    }

    if (delimiter != NULL)
        *delimiter = aChar;

    *args = workArgs;
    return isArg;
}

