#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sched.h>
#include <errno.h>

#include "tsadclib1624.h"

/* Possible improvements:
 * - expand for use with TS-ADC24
 * - add support for TS-8100
 *
 * 20100311 jwl - Corrections to the ADC24 12 bit and twos complement logic
 *
 * 20100415 jwl - Correction to tsadclib1624_moveData to properly
 *                handle non-blocking mode
 *
 */

#define CUSHION 16
#define WAITCOUNT 128

bool allBoardsConfigured = false;
bool tsadclib1624_verboseFlag = false;

tsadcinstance allAdc;
boarddata *anAdc;

uint8_t *tsadclib1624ErrorText[] = {
    "Can't openmem",
    "Can't mapio16",
    "Can't mapio8",
    "Not initialized",
    "Bad hardware",
    "Bad parameter",
    "Bad board",
    "Can't open IRQ6",
    "Can't open IRQ7",
    "FIFO Overrun"
};

uint8_t *tsadclib1624_errors(int errorCode)
{
    if (errorCode <= TSADCERR_MAXERROR)
        return "";
    else
        return tsadclib1624ErrorText[abs(errorCode) - 1];
}

void tsadclib1624_verbose(bool Verbose)
{
    tsadclib1624_verboseFlag = Verbose;
}

static void tsadclib1624_dumpBoardConfig(boarddata *theAdc)
{
    if (theAdc == NULL)
    {
        fprintf(stderr, "dumpBoardConfig: Invalid board address\n");
        return;
    }

    fprintf(stderr, "\ndumpBoardConfig: operational:%s  IRQ%c Twos Comp:%s  ADC%s\n",
            theAdc->operational ? "Yes" : "No",
            theAdc->irq7 ? '7' : '6',
            theAdc->twosComplementResult ? "Yes" : "No",
            theAdc->adc16 ? "16" : "24");

    fprintf(stderr, "dumpBoardConfig: Channel A:%d %d @ %p  B:%d %d @ %p\n",
            theAdc->chana, theAdc->cnt1, theAdc->data1,
            theAdc->chanb, theAdc->cnt2, theAdc->data2);

    fprintf(stderr, "dumpBoardConfig: SampleCnt:%d sSampleCnt:%d wSampleCnt:%d\n",
            theAdc->sampleCnt, theAdc->sSampleCnt, theAdc->wSampleCnt);

    fprintf(stderr, "dumpBoardConfig: PORT Cnt1:%d  Cnt2:%d Sample Cnt:%d cArray[%p] dArray[%p]\n",
            theAdc->sData1, theAdc->sData2, theAdc->sSampleCnt, allAdc.fifoChannelBasePtr, allAdc.fifoDataBasePtr);

    fprintf(stderr, "dumpBoardConfig: ADC Board:0x%02x (%s) Acquisition Mode:%s io8:%p io16:%p\n",
            theAdc->whichadcboard, theAdc->whichadcboard == 0x3e ? "Adc16" : "Adc24",
            theAdc->howAcquire ? "Aqcuire" : "Sample", theAdc->io8, theAdc->io16);
}

/*
 *   tsadclib1624_buffers
 *
 *   Return true if biffers assigned or false if not
 */

bool tsadclib1624_buffers(void)
{
    if (allAdc.fifoChannelBasePtr == NULL ||
            allAdc.fifoDataBasePtr == NULL)
        return false;
    else
        return true;
}

static void fixupBuffer(boarddata *theAdc)
{
    int32_t  totalSampleCount = theAdc->sampleCnt;
    uint16_t *dPtr = allAdc.fifoDataBasePtr;

    if (theAdc->whichadcboard == TS_ADC16)
        return;

    while (totalSampleCount--)
    {
        *dPtr &= 0x0fff;  //  Strip off the channel information

        if (theAdc->twosComplementResult)
            *dPtr = (((*dPtr) ^ 0x0fff) + 1);  // two complement 12-bit result

        dPtr++;
    }
}


/*
 * adclibts1624_init - Initializes the adc.
 *
 * On success the adc is usable and a zero (0) is returned.
 * On failure a negative integer indicating the fault is returned.
 * (The fault values are defined in header file adclibts1624.h)
*/

int tsadclib1624_init(sbctype sbc, uint16_t maxChannel, uint32_t maxBufferSize)
{
    uint8_t jumpers;
    unsigned int offset;
    short index;
    uint8_t boardType;
    bool foundBoard = false;
    bool foundAdc24Board = false;


    if (allBoardsConfigured)  //  Don't do it again
        return 0;

    switch (sbc)
    {
        case SBC_TS7200:
            allAdc.IoBase8  = 0x11E00000;
            allAdc.IoBase16 = 0x21E00000;
            allAdc.IoOffset = 0x100;
            allAdc.theSbc = SBC_TS7200;
            if (tsadclib1624_verboseFlag)
                fprintf(stderr, "tsadclib1624_init: TS7200\n");
            break;

        case SBC_TS7350:
            allAdc.IoBase8  = 0x600FF000;
            allAdc.IoBase16 = 0x600FF000;
            allAdc.IoOffset = 0x900;
            allAdc.theSbc = SBC_TS7350;
            if (tsadclib1624_verboseFlag)
                fprintf(stderr, "tsadclib1624_init: TS7350/TS7370\n");
            break;

        case SBC_TS7800:
            allAdc.IoBase8  = 0xEE000000;
            allAdc.IoBase16 = 0xEF000000;
            allAdc.IoOffset = 0x100;
            allAdc.theSbc = SBC_TS7800;
            if (tsadclib1624_verboseFlag)
                fprintf(stderr, "tsadclib1624_init: TS7800\n");
            break;

        default:
            return TSADCERR_BADPARAMETER;
    }

    /*
     *   Set allAdc.Io16 as a pointer into memory space.
     */

    if ((allAdc.memoryFd = open("/dev/mem", O_RDWR|O_SYNC)) == -1)
        return TSADCERR_CANTOPENMEM;

    allAdc.Io16 = (uint16_t *) mmap(0, getpagesize(),
            PROT_READ|PROT_WRITE, MAP_SHARED, allAdc.memoryFd, allAdc.IoBase16);
    if (allAdc.Io16 == MAP_FAILED)
        return TSADCERR_CANTMAPIO16;

    allAdc.Io8  = (uint8_t *) mmap(0, getpagesize(),
            PROT_READ|PROT_WRITE, MAP_SHARED, allAdc.memoryFd, allAdc.IoBase8);
    if (allAdc.Io8 == MAP_FAILED)
        return TSADCERR_CANTMAPIO8;

#ifdef USING_IRQS
    allAdc.Irq6FileHandle = 0;
    allAdc.Irq7FileHandle = 0;
#endif

    /* Configure PC104 I/O if required */
    if (sbc == SBC_TS7350)
    {
        /* Configure PC/104 connector on this SBC for PC/104: */
        CONFIG(0x98) = 0x5555;
        CONFIG(0x9a) = 0x5555;
        CONFIG(0x9c) = 0x5555;
        CONFIG(0x9e) = 0x5555;
        /* Configure ISA bus for 16 bit ARM mode: */
        CONFIG(0xd4) |= 0x0800;
    }
    else if (sbc == SBC_TS7800)
    {
        /* Configure PC/104 connector on this SBC for PC/104: */
        CONFIG(0x30) = 0x5555;
        CONFIG(0x32) = 0x5555;
        CONFIG(0x34) = 0x5555;
        CONFIG(0x36) = 0x5555;
        /* Configure ISA bus for 16 bit ARM mode: */
        CONFIG(0x0c) |= 0x0800;
    }
    else if (sbc != SBC_TS7200)
    {
        fprintf(stderr, "Invalid sbc ID: 0x%02x.\n", sbc);
        return TSADCERR_BADPARAMETER;
    }

    // Scan for installed boards and setup access array parameters
    for (index = 0, offset = 0x00; index < 4; index++, offset += BOARD_SPACING)
    {
        anAdc = &allAdc.boards[index];
        anAdc->io16 = allAdc.Io16 + ((allAdc.IoOffset + offset)/sizeof(uint16_t));
        anAdc->io8  = allAdc.Io8 + allAdc.IoOffset + offset;
        anAdc->operational = 1;
        boardType = ADC8(anAdc, BOARDID_INDEX);
        jumpers = ADC8(anAdc, BOARDID_INDEX + 1);

        if (tsadclib1624_verboseFlag)
            fprintf(stderr, "tsadclib1624_init Io16:%p Io8:%p io16:%p io8:%p Index:0x%02x ID:0x%02x\n",
                    allAdc.Io16, allAdc.Io8, anAdc->io16, anAdc->io8, offset, boardType);

        if (boardType == TS_ADC16)
        {
            anAdc->whichadcboard = TS_ADC16;
            anAdc->adc16 = 1;
            foundBoard = true;
            if (tsadclib1624_verboseFlag)
                fprintf(stderr, "ADC16 found at offset 0x%02x ", offset);
        }
        else if (boardType == TS_ADC24)
        {
            anAdc->whichadcboard = TS_ADC24;
            anAdc->adc16 = 0;
            foundAdc24Board = true;
            foundBoard = true;
            if (tsadclib1624_verboseFlag)
                fprintf(stderr, "ADC24 found at offset 0x%02x ", offset);
        }
        else
        {
            anAdc->whichadcboard = TS_ADCINVALID;
            anAdc->operational = 0;
        }

        if (anAdc->operational)
        {
            if (tsadclib1624_verboseFlag)
                fprintf(stderr, "jumpers: (0x%02x) %s %s %s %s (%s)\n", jumpers,
                        jumpers & JP1 ? "JP1" : "", jumpers & JP2 ? "JP2" : "",
                        jumpers & JP3 ? "JP3" : "", jumpers & JP4 ? "JP4" : "",
                        (jumpers & JP4) ? "IRQ7" : "IRQ6");

            if (!(jumpers & JP3)) /* JP3 must be ON for ARM */ 
            {
                if (tsadclib1624_verboseFlag)
                    fprintf(stderr, "JP3 is NOT set and is required for ARM\n");
                return TSADCERR_HARDWARE;
            }

            // ISA IRQs 6 and 7 map to Linux IRQs and are
            // detected by JP4 = 0 & JP4 = 1 respectively
            // TS7800 69 & 70 respectively
            // EP9302 33 & 40 respectively
        }
    }

    if (sbc != SBC_TS7200)
    {
#ifdef USING_IRQS
        //  Save this boards interrupt number
        anAdc->irq7 = (jumpers & JP4) ? 1 : 0;

        //  Setup interrupts /proc access string for 7800 vs. 7200 & 7350
        sprintf(allAdc.irqstring6, "/proc/irq/%s/irq", (sbc == SBC_TS7800) ? "69" : "33");
        sprintf(allAdc.irqstring7, "/proc/irq/%s/irq", (sbc == SBC_TS7800) ? "70" : "40");

        allAdc.Irq6FileHandle = open(allAdc.irqstring6, O_RDONLY);
        if (allAdc.Irq6FileHandle < 0)
        {
            fprintf(stderr, "Failed to open \'%s\' errno: %d  %s\n", allAdc.irqstring6, errno, strerror(errno));
            return TSADCERR_CANTOPENIRQ6;
        }

        allAdc.Irq7FileHandle = open(allAdc.irqstring7, O_RDONLY);
        if (allAdc.Irq7FileHandle < 0)
        {
            fprintf(stderr, "Failed to open \'%s\' errno: %d  %s\n", allAdc.irqstring7, errno, strerror(errno));
            return TSADCERR_CANTOPENIRQ7;
        }
#endif
    }

    allAdc.theMaxChannelRequest = maxChannel > 24 ? 24 : maxChannel;
    allAdc.theMaxBufferSize = maxBufferSize;

    if (foundBoard)
    {
        if (!foundAdc24Board && maxChannel > 16)
            allAdc.theMaxChannel = 16;
        else
            allAdc.theMaxChannel = maxChannel;

        allAdc.theTotalBufferSize = allAdc.theMaxChannel * allAdc.theMaxBufferSize;

        if (allAdc.theTotalBufferSize)
        {
            allAdc.fifoDataBaseCnt = allAdc.theTotalBufferSize * sizeof(uint16_t);
            if ((allAdc.fifoDataBasePtr = malloc(allAdc.fifoDataBaseCnt)) == NULL)
            {
                fprintf(stderr, "tsadclib1624_init: Failed data buffer allocation: %d\n", allAdc.theTotalBufferSize * sizeof(uint16_t));
                return TSADCERR_OUTOFMEMORY;
            }

            allAdc.fifoChannelBaseCnt = allAdc.theTotalBufferSize * sizeof(uint8_t);
            if ((allAdc.fifoChannelBasePtr = malloc(allAdc.fifoChannelBaseCnt)) == NULL)
            {
                fprintf(stderr, "tsadclib1624_init: Failed channel buffer allocation: %d\n", allAdc.theTotalBufferSize * sizeof(uint8_t));
                return TSADCERR_OUTOFMEMORY;
            }
        }
        else
        {
            allAdc.fifoChannelBasePtr = NULL;
            allAdc.fifoDataBasePtr = NULL;
        }
    }
    else
        return TSADCERR_BADBOARD;

    allBoardsConfigured = true;

    return 0;
}

/*
 *  tsadclib1624_boardstatus
 *  return the status of a board/adc based on the board
 *  index (set by JP1 & JP2, range 0-3.
 */
adctype tsadclib1624_boardstatus(BoardIndex theBoard)
{
    if (!allBoardsConfigured)
        return TSADCERR_NOINIT;
    if (theBoard >= ADC_BOARDMAX)
        return TSADCERR_BADBOARD;
    anAdc = &allAdc.boards[theBoard];

    return anAdc->whichadcboard;
}

int tsadclib16_digitalOut(BoardIndex theBoard, bool state)
{
    if (theBoard >= ADC_BOARDMAX)
        return TSADCERR_BADBOARD;
    anAdc = &allAdc.boards[theBoard];
    if (!anAdc->operational)
        return TSADCERR_NOINIT;
    if (!anAdc->adc16)
        return TSADCERR_BADPARAMETER;

    /* set or clear digital output bit: */
    if (state)
        ADC(anAdc, DIO_REG) |= DIO_REG_BIT;
    else
        ADC(anAdc, DIO_REG) &= ~DIO_REG_BIT;

    return 0;
}

uint8_t tsadclib16_digitalIn(BoardIndex theBoard)
{
    if (theBoard >= ADC_BOARDMAX)
        return TSADCERR_BADBOARD;
    anAdc = &allAdc.boards[theBoard];
    if (!anAdc->operational)
        return TSADCERR_NOINIT;
    if (!anAdc->adc16)
        return TSADCERR_BADPARAMETER;

    return ADC(anAdc, DIO_REG) & 0xf;
}

uint16_t tsadclib16_counterIn(BoardIndex theBoard, CounterIndex theCounter)
{
    if (theBoard >= ADC_BOARDMAX)
        return TSADCERR_BADBOARD;
    anAdc = &allAdc.boards[theBoard];
    if (!anAdc->operational)
        return TSADCERR_NOINIT;
    if (!anAdc->adc16 || theCounter >= ADC_COUNTERMAX)
        return TSADCERR_BADPARAMETER;

    return (ADC(anAdc, (COUNTER_REG_BASE + (theCounter * 2))));
}

int16_t tsadclib16_dacset(BoardIndex theBoard, uint8_t theDac, dacrefrange referenceRange, uint16_t value)
{
    if (theBoard >= ADC_BOARDMAX)
        return TSADCERR_BADBOARD;
    anAdc = &allAdc.boards[theBoard];
    if (!anAdc->operational)
        return TSADCERR_NOINIT;
    if (!anAdc->adc16 || theDac >= DACMAX)
        return TSADCERR_BADPARAMETER;

    switch (referenceRange)
    {
        default:
            return TSADCERR_BADPARAMETER;

        case DACRANGE_0VU:
        case DACRANGE_0VB:
        case DACRANGE_02VU:
        case DACRANGE_02VB:
            break;
    }

    value |= (((uint16_t) theDac) << 14);
    value |= referenceRange;

    ADC(anAdc, DAC_REG_BASE) = value;
    ADC8(anAdc, DAC_REG_BASE + 1) = value >> 8;

    return 0;
}

/*
 * tsadclib1624_startAcquisution will acquire sampleCnt samples at sampleRate Hz.
 * The acquisition is started and the function returns
 *
 * It will return the number of samples actually acquired.
*/
int16_t tsadclib1624_startAcquisution(BoardIndex theBoard, int8_t channel, uint32_t sampleRate, int16_t sampleCnt, adcrange range, uint16_t *data1, uint16_t *data2, bool contineOperation)
{
    unsigned int pacing;
    uint16_t adccfg;
    uint16_t numChan;

    if (theBoard >= ADC_BOARDMAX)
        return TSADCERR_BADBOARD;
    anAdc = &allAdc.boards[theBoard];
    if (!anAdc->operational)
        return TSADCERR_NOINIT;

    if ((anAdc->whichadcboard == TS_ADC16 && channel > 15) ||
            (anAdc->whichadcboard == TS_ADC24 && channel > 23))
        return TSADCERR_BADPARAMETER;

    if (channel > allAdc.theMaxChannel)
        return TSADCERR_BADPARAMETER;

    //   Flush the FIFO
    if (!contineOperation)
    {
        ADC(anAdc, FPGA_CONFIGURATION) &= ~FPGA_RUNNING; // stop ADC machine
        while (FIFO_COUNT(anAdc))    // And flush the FIFO
            ADC(anAdc, FIFO_READ);

        //  Specuak case, where caller want to stop operation of the board
        if (sampleCnt == 0)
        {
            anAdc->chana = 0;
            anAdc->chanb = 0;
            anAdc->sampleCnt = 0;
            anAdc->cnt1 = 0;
            anAdc->cnt2 = 0;
            anAdc->data1 = NULL;
            anAdc->data2 = NULL;
            return 0;
        }
    }

    anAdc->howAcquire = ADC_ACQUIRE;
    anAdc->chana = channel & ~1;
    //  If no buffer for data2, then don't attaempt collection
    //  If channel B is set at maximum or above, we save checking
    //  addresses in moveData
    //  Ensure channel B never matvches if data2 is NULL
    anAdc->chanb = data2 != NULL ? ((channel & ~1) + 1) : 24;
    anAdc->cnt1 = 0;
    anAdc->cnt2 = 0;

    //  Setup actual sample count
    //  If ADC16 we take sample pairs;  ADC24 takes sample quads
    //  Calculate the hardware sample channels and then calculate proper number
    //  of samples to get desired samples on the channel(s) of interest
    if (anAdc->adc16)
    {
        numChan = (uint16_t) ((channel & ~1) >> 1);
        anAdc->sampleCnt = sampleCnt * ((numChan << 1) + 2);
    }
    else
    {
        numChan = (uint16_t) ((channel & ~3) >> 2);
        anAdc->sampleCnt = sampleCnt * ((numChan + 1) * 4);
    }

    anAdc->data1 = data1;
    anAdc->data2 = data2;
    anAdc->sSampleCnt = sampleCnt;
    anAdc->wSampleCnt = 0;
    anAdc->sData1 = data1;
    anAdc->sData2 = data2;

    /*
     * Determine if the 12 bit adc output is binary or two's complement
     * The way this is determined is not obvious, but is detailed
     * in the AD7266 data sheet on page 11, using tables 5 and 6
     */

    if (!anAdc->adc16)  //  If board is ADC-24
    {
        if ((range == ADCRANGE_0VS) || (range == ADCRANGE_0VD && (channel & 2)))
            anAdc->twosComplementResult = false;
        else
            anAdc->twosComplementResult = true;
    } 


    /* set config for all channels up to channel */
    adccfg = range | (uint16_t) (numChan << 1);
    ADC(anAdc, FPGA_CONFIGURATION) = FPGA_RUNNING; /* start and stop to reset ADC16 */
    ADC(anAdc, FPGA_CONFIGURATION) = adccfg;
    while (FIFO_COUNT(anAdc))
        ADC(anAdc, FIFO_READ);
    pacing = FPGA_CLOCK/sampleRate;
    if (pacing < 0x140)
        pacing = 0;
    ADC(anAdc, DELAY_LSB) = pacing & 0xffff;
    ADC(anAdc, DELAY_MSB) = pacing >> 16;
    ADC(anAdc, FPGA_CONFIGURATION) = adccfg;

    if (tsadclib1624_verboseFlag)
    {
        fprintf(stderr, "sampleCnt=%d\n", anAdc->sampleCnt);
        fprintf(stderr, "pacing=0x%x\n", pacing);
        fprintf(stderr, "adccfg=0x%x\n", ADC(anAdc, FPGA_CONFIGURATION));
    }

    /* Indicate need to start ADC machine: */
    allAdc.startAcquisition = 1;
    return 0;
}

/*
 *   Separate data from the main buffer into the supplied buffer(s)
 *
 *   NOTE: This function only knows how to move data.  It is called from the 
 *   function tsadclib1624_pollResults with all reauired parameters to fetch
 *   the cirrently available data.  It knows how and where to place the data
 *   and when to stop acquiring the data
 */

int32_t tsadclib1624_seperateData(boarddata *theAdc)
{
    int32_t  totalSampleCount = theAdc->sampleCnt;
    int32_t  channelSampleCount = theAdc->sSampleCnt;
    uint8_t  *cPtr = allAdc.fifoChannelBasePtr;
    uint16_t *dPtr = allAdc.fifoDataBasePtr;
    uint16_t *data1 = theAdc->data1;
    uint16_t *data2 = theAdc->data2;
    uint16_t index;
    uint16_t theChannel;
    int16_t theData;
    int16_t **dataArray;
    uint32_t channelCounters[24];

    if (anAdc->howAcquire == ADC_ACQUIRE)
    {
        printf("seperateData: ADC_ACQUIRE\n");
        while (totalSampleCount--)
        {
            if (*cPtr == theAdc->chana && theAdc->cnt1 < theAdc->sSampleCnt)
            {
                *data1++ = *dPtr;
                theAdc->cnt1++;
            }
            else if (*cPtr == theAdc->chanb && theAdc->cnt2 < theAdc->sSampleCnt)
            {
                *data2++ = *dPtr;
                theAdc->cnt2++;
            }
            cPtr++;
            dPtr++;
        }
        return theAdc->cnt1;
    }
    else if (anAdc->howAcquire == ADC_SAMPLE)
    {
        for (index = 0; index < 24; index++)
            channelCounters[index] = 0;

        dataArray = theAdc->arrayData;
        printf("seperateData: ADC_SAMPLE\n");
        while (totalSampleCount--)
        {
            theChannel = *cPtr++;
            if (dataArray[theChannel] != NULL)
                dataArray[theChannel][channelCounters[theChannel]++] = *dPtr++;
        }
        return channelCounters[0];
    }
}

/*
 *   Move data from the FIFO into the supplied buffer(s)
 *
 *   NOTE: This function only knows how to move data.  It is called from the 
 *   function tsadclib1624_pollResults with all required parameters to fetch
 *   the cirrently available data.  It knows how and where to place the data
 *   and when to stop acquiring the data
 *
 *   Returns: -value is error code
 *            zero   all data not yet available
 *            +value Requested data available in the buffer
 *                   supplied to: tsadclib1624_startAcquisution
 */

static int32_t tsadclib1624_moveData(boarddata *theAdc)
{
    uint16_t fifocount;

    while (ADC(theAdc, FPGA_CONFIGURATION) & FPGA_RUNNING)
    {
        if ((fifocount = FIFO_COUNT(theAdc)) <= CUSHION)
            break;

        while (fifocount > CUSHION)
        {
            if (anAdc->wSampleCnt < anAdc->sampleCnt)
            {
                *allAdc.fifoChannelWorkPtr++ = FIFO_CHANNEL(anAdc);
                *allAdc.fifoDataWorkPtr++    = ADC(theAdc, FIFO_READ);
                allAdc.fifoWorkCnt++;
                anAdc->wSampleCnt++;
                fifocount--;
            }
            else
            {
                ADC(theAdc, FPGA_CONFIGURATION) &= ~FPGA_RUNNING;
                fprintf(stderr, "Complete transfer: ADC Machine PROGRAM STOP: 0x%04x\n", ADC(theAdc, FPGA_CONFIGURATION));
                return anAdc->sampleCnt;
            }
        }
    }

    //  Check for FIFO overrun
    if (!(ADC(theAdc, FPGA_CONFIGURATION) & FPGA_RUNNING))
    {
        fprintf(stderr, "ADC machine stopped prematurely due to FIFO overflow: 0x%04x Fifo: %d\n",
                ADC(theAdc, FPGA_CONFIGURATION), fifocount);
        return TSADCERR_FIFOOVERRUN;;
    }
    else
    {
        return TSADCSTATUS_NODATA;
    }
}

/*
 *   tsadclib1624_pollResults
 *   Parameters:  The board (BOARD0 - BOARD3) to get the data from
 *                Blocking or NonBlocking mode flag
 *   Returns: -value is error code
 *            zero   no data available yet
 *            +value Requested data available in the buffer
 *                   supplied to: tsadclib1624_startAcquisution
 *   NOTE: This function only knows how to waut for data available
 *   for all supported SBC/kernel versions in blocking and non-blocking modes
 */

int32_t tsadclib1624_pollResults(BoardIndex theBoard, bool blocking)
{
    uint16_t fifocount;
    boarddata *theAdc;
    int32_t rtnStatus;
#ifdef USING_IRQS
    int IrqFileHandle;
#endif
    fd_set FdSet;
    struct timeval timeValue;
    struct sched_param oldSched, newSched;
    uint16_t waitcount = WAITCOUNT;

    if (anAdc->sampleCnt < WAITCOUNT)
        waitcount = anAdc->sampleCnt;

    if (theBoard >= ADC_BOARDMAX)
        return TSADCERR_BADBOARD;
    theAdc = &allAdc.boards[theBoard];
    if (!theAdc->operational)
        return TSADCERR_NOINIT;

    sched_getparam(0, &oldSched);
    sched_getparam(0, &newSched);
    newSched.sched_priority = sched_get_priority_max(SCHED_FIFO); 

    if (allAdc.startAcquisition)
    {
        ADC(anAdc, FPGA_CONFIGURATION) |= FPGA_RUNNING;  //  Start scanning
        allAdc.startAcquisition = 0;
        if (anAdc->sampleCnt > allAdc.theTotalBufferSize)
            anAdc->sampleCnt = allAdc.theTotalBufferSize;
        allAdc.fifoWorkCnt  = 0;
        allAdc.fifoChannelWorkPtr = allAdc.fifoChannelBasePtr;
        allAdc.fifoDataWorkPtr = allAdc.fifoDataBasePtr;
    }

    fifocount = FIFO_COUNT(theAdc);

    if (!blocking)
    {
        int32_t status;

        status = tsadclib1624_moveData(theAdc);
        if (status > 0)
        {
            fixupBuffer(theAdc);  //  Fixup buffer if ADC24 (12 bit ADC)
            return tsadclib1624_seperateData(theAdc);
        }
        else
            return status;
    }
    else  // Process blocking
    {
        rtnStatus = 0;
        while (rtnStatus == 0)
        {
            rtnStatus = tsadclib1624_moveData(theAdc);
            if (rtnStatus < 0)
                return rtnStatus;

#ifdef USING_IRQS
            if (allAdc.theSbc != SBC_TS7200)
                while (select(IrqFileHandle+1, &FdSet, NULL, NULL, &timeValue));
#endif
            if (ADC(theAdc, FPGA_CONFIGURATION) & FPGA_RUNNING) //  If machine still Running
            {
                while (fifocount < waitcount)
                {
                    sched_setscheduler(0, SCHED_FIFO, &newSched);
                    //
                    //  At maximum aggretae rate the sample time is 4 us per sample
                    //  4 * 128 samples (1/4 of the FIFO) is 512 us, so we wait 500
                    //
fputc('.', stderr); fflush(stderr);
                    usleep(500);
                    sched_setparam(0, &oldSched);
                    fifocount = FIFO_COUNT(theAdc);
                }
            }
        }
    }
    fixupBuffer(theAdc);  //  Fixup buffer if ADC24 (12 bit ADC)
    return tsadclib1624_seperateData(theAdc);
}

/*
 *  tsadclib1624_sample
 *
 *  Acquire numberOfSamples samples from all channels on ADC theBoard from zero
 *  through maxChannel in range range at rate sampleRate into buffer array
 *  channelBuffer[0...maxChannel]
 */
int16_t tsadclib1624_sample(BoardIndex theBoard, int16_t maxChannel, int32_t sampleRate, int16_t sampleCnt, adcrange range, int16_t *channelBuffers[], bool contineOperation)
{
    unsigned int pacing;
    uint16_t adccfg;
    uint16_t numChan;

    uint16_t loop;

#ifdef NEVER
    if (channelBuffers == NULL)
    {
        fprintf(stderr, "sample: %d @ %p\n", sampleCnt, channelBuffers);
    }
    else
    {
        for (loop = 0; loop < 24; loop++)
        {
            if ((loop % 8) == 0)
                fprintf(stderr, "\n");
            fprintf(stderr, "%p  ", channelBuffers[loop]);
        }
        fprintf(stderr, "\n");
    }
#endif

    if (theBoard >= ADC_BOARDMAX)
        return TSADCERR_BADBOARD;
    anAdc = &allAdc.boards[theBoard];
    if (!anAdc->operational)
        return TSADCERR_NOINIT;

    if ((anAdc->whichadcboard == TS_ADC16 && maxChannel > 15) ||
            (anAdc->whichadcboard == TS_ADC24 && maxChannel > 23))
        return TSADCERR_BADPARAMETER;

    if (maxChannel > allAdc.theMaxChannel)
        return TSADCERR_BADPARAMETER;

    //   Flush the FIFO
    if (!contineOperation)
    {
        ADC(anAdc, FPGA_CONFIGURATION) &= ~FPGA_RUNNING; // stop ADC machine
        while (FIFO_COUNT(anAdc))    // And flush the FIFO
            ADC(anAdc, FIFO_READ);

        //  Specuak case, where caller want to stop operation of the board
        if (sampleCnt == 0)
        {
            anAdc->chana = 0;
            anAdc->chanb = 0;
            anAdc->sampleCnt = 0;
            anAdc->cnt1 = 0;
            anAdc->cnt2 = 0;
            anAdc->data1 = NULL;
            anAdc->data2 = NULL;
            return 0;
        }
    }

    anAdc->howAcquire = ADC_SAMPLE;
    anAdc->chana = 24;
    anAdc->chanb = 24;
    anAdc->cnt1 = 0;
    anAdc->cnt2 = 0;


    //  Setup actual sample count
    //  If ADC16 we take sample pairs;  ADC24 takes sample quads
    //  Calculate the hardware sample channels and then calculate proper number
    //  of samples to get desired samples on the channel(s) of interest
    if (anAdc->adc16)
    {
        numChan = (uint16_t) ((maxChannel & ~1) >> 1);
        anAdc->sampleCnt = sampleCnt * ((numChan << 1) + 2);
        //fprintf(stderr, "sample 16: channel: %d  numChan:%d  sample:%d  total:%d\n", maxChannel, numChan, sampleCnt, anAdc->sampleCnt);
    }
    else
    {
        numChan = (uint16_t) ((maxChannel & ~3) >> 2);
        anAdc->sampleCnt = sampleCnt * ((numChan + 1) * 4);
        //fprintf(stderr, "sample 24: channel: %d  numChan:%d  sample:%d  total:%d\n", maxChannel, numChan, sampleCnt, anAdc->sampleCnt);
    }

    anAdc->data1 = NULL;
    anAdc->data2 = NULL;
    anAdc->sSampleCnt = sampleCnt;
    anAdc->wSampleCnt = 0;
    anAdc->sData1 = NULL;
    anAdc->sData2 = NULL;
    anAdc->arrayData = channelBuffers;

    /*
     * Determine if the 12 bit adc output is binary or two's complement
     * The way this is determined is not obvious, but is detailed
     * in the AD7266 data sheet on page 11, using tables 5 and 6
     */

    if (!anAdc->adc16)
    {
        if ((range == ADCRANGE_0VS) || (range == ADCRANGE_02VD && (maxChannel & 1)))
            anAdc->twosComplementResult = false;
        else
            anAdc->twosComplementResult = true;
    }

    /* set config for all channels up to channel */
    adccfg = range | (uint16_t) (numChan << 1);
    ADC(anAdc, FPGA_CONFIGURATION) = FPGA_RUNNING; /* start and stop to reset ADC16 */
    ADC(anAdc, FPGA_CONFIGURATION) = adccfg;
    while (FIFO_COUNT(anAdc))
        ADC(anAdc, FIFO_READ);
    pacing = FPGA_CLOCK/sampleRate;
    if (pacing < 0x140)
        pacing = 0;
    //fprintf(stderr, "sample - pacing:%d numChan:%d samples:%d AdcCfg:%03x\n", pacing, numChan, anAdc->sampleCnt, adccfg);
    ADC(anAdc, DELAY_LSB) = pacing & 0xffff;
    ADC(anAdc, DELAY_MSB) = pacing >> 16;
    ADC(anAdc, FPGA_CONFIGURATION) = adccfg;
    if (tsadclib1624_verboseFlag)
    {
        fprintf(stderr, "sampleCnt=%d\n", anAdc->sampleCnt);
        fprintf(stderr, "pacing=0x%x\n", pacing);
        fprintf(stderr, "adccfg=0x%x\n", ADC(anAdc, FPGA_CONFIGURATION));
    }

    /* Indicate need to start ADC machine: */
    allAdc.startAcquisition = 1;
    return 0;
}

